/*
*   Copyright 2007 by Alex Merry <alex.merry@kdemail.net>
*   Copyright 2008 by Alexis Ménard <darktears31@gmail.com>
*   Copyright 2011 by Eugeni Kozhanov <eugeni.kozhanov@rosalab.ru>
*
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU Library General Public License version 2,
*   or (at your option) any later version.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details
*
*   You should have received a copy of the GNU Library General Public
*   License along with this program; if not, write to the
*   Free Software Foundation, Inc.,
*   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "panel.h"

#include <limits>

#include <QDesktopWidget>
#include <QAction>
#include <QApplication>
#include <QGraphicsLinearLayout>
#include <QGraphicsLayout>
#include <QGraphicsSceneDragDropEvent>
#include <QVBoxLayout>
#include <QTimer>
#include <QPainter>
#include <QDir>

#include <KDebug>
#include <KIcon>
#include <KIconLoader>

#include <Plasma/Corona>
#include <Plasma/FrameSvg>
#include <Plasma/Theme>
#include <Plasma/AbstractToolBox>
#include <Plasma/View>
#include <Plasma/PaintUtils>
#include <Plasma/WindowEffects>

//#include <kephal/screens.h>
#include <Plasma/Separator>

#include "configpanel.h"

#include <kio/copyjob.h>
#include <kio/netaccess.h>
#include <KDesktopFile>

using namespace Plasma;

int DEFAULT_PANEL_SIZE = 50;

class Spacer : public QGraphicsWidget
{
public:
    Spacer(QGraphicsWidget *parent)
         : QGraphicsWidget(parent),
           m_visible(true)
    {
        setAcceptDrops(true);
    }

    ~Spacer()
    {}

    Panel *panel;
    bool m_visible;

protected:
    void dropEvent(QGraphicsSceneDragDropEvent *event)
    {
        event->setPos(mapToParent(event->pos()));
        panel->dropEvent(event);
    }

    void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget * widget = 0)
    {
        Q_UNUSED(option)
        Q_UNUSED(widget)

        if (!m_visible) {
            return;
        }

	/*
	Plasma::Applet::List pleftApplets = containment()->applets();
        foreach(Plasma::Applet* ap, pleftApplets){
                        if(ap->pluginName() == "rosaicon"){
			}
	}
	*/

        //TODO: make this a pretty gradient?
        painter->setRenderHint(QPainter::Antialiasing);
        QPainterPath p = Plasma::PaintUtils::roundedRectangle(contentsRect().adjusted(1, 1, -2, -2), 4);
        QColor c = Plasma::Theme::defaultTheme()->color(Plasma::Theme::TextColor);
        c.setAlphaF(0.7);

        painter->fillPath(p, c);
    }
};

Panel::Panel(QObject *parent, const QVariantList &args)
    : Containment(parent, args),
      m_configureAction(0),
      m_currentSize(QSize(DEFAULT_PANEL_SIZE, DEFAULT_PANEL_SIZE)),
      m_maskDirty(true),
      m_canResize(false),
      m_spacerIndex(-1),
      m_spacer(0),
      m_lastSpace(0),
      m_layout(0),
      m_resizedApplets(0),
      m_mainConfiguration(0)
{
    setContainmentType(Containment::PanelContainment);
    setDrawWallpaper(true);

    printf("[*]Rosa panel init!\n");

    m_background = new Plasma::FrameSvg(this);
    m_background->setImagePath("widgets/panel-background");
    m_background->setEnabledBorders(Plasma::FrameSvg::AllBorders);
    connect(m_background, SIGNAL(repaintNeeded()), this, SLOT(backgroundChanged()));

    m_lastSpaceTimer = new QTimer(this);
    m_lastSpaceTimer->setSingleShot(true);
    connect(m_lastSpaceTimer, SIGNAL(timeout()), this, SLOT(adjustLastSpace()));

    m_enableUpdateResizeTimer = new QTimer(this);
    m_enableUpdateResizeTimer->setSingleShot(true);
    m_enableUpdateResizeTimer->setInterval(100);
    connect(m_enableUpdateResizeTimer, SIGNAL(timeout()), this, SLOT(enableUpdateSize()));

    m_updateSizeTimer = new QTimer(this);
    m_updateSizeTimer->setSingleShot(true);
    m_updateSizeTimer->setInterval(10);
    connect(m_updateSizeTimer, SIGNAL(timeout()), this, SLOT(updateSize()));
    
    m_initTimer = new QTimer(this);
    m_initTimer->setSingleShot(true);
    //m_initTimer->setInterval(3000);
    connect(m_initTimer, SIGNAL(timeout()), this, SLOT(initRosa()));

    connect(this, SIGNAL(appletRemoved(Plasma::Applet*)), this, SLOT(appletWasRemoved(Plasma::Applet*)));

    m_initTimer->start(700);

    m_sizeCheckTimer = new QTimer(this);
    connect(m_sizeCheckTimer, SIGNAL(timeout()), this, SLOT(sizeCheck()));
    m_sizeCheckTimer->start(500);
    m_sizeCheckCounter = 20;

}

Panel::~Panel()
{
}

void Panel::sizeCheck()
{
    QPointF curPos = pos();
    curPos.setX(curPos.x() + 1);
    setPos(curPos);

    m_sizeCheckCounter--;

    if(!m_sizeCheckCounter)
	m_sizeCheckTimer->stop();
}

void Panel::init()
{
    Containment::init();
    //FIXME: This should be enabled, but in that case proxywidgets won't get rendered
    //setFlag(ItemClipsChildrenToShape, true);
    //START ROSA patch
    QRect screenRect = QDesktopWidget().screenGeometry();

    //if (formFactor() == Plasma::Vertical) {
	//    DEFAULT_PANEL_SIZE = screenRect.width()/20;
    //}else{
	    DEFAULT_PANEL_SIZE = screenRect.height()/20;
    //}

    m_layout = new QGraphicsLinearLayout(this);
    m_layout->setSpacing(2);
    //m_layout->addStretch(1);
    m_layout->setSizePolicy(QSizePolicy(QSizePolicy::Expanding,QSizePolicy::Expanding));
    updateBorders(geometry().toRect());
    setLayout(m_layout);
    m_layout->setMaximumSize(size());

    KConfigGroup cg = config("Configuration");

    m_currentSize = cg.readEntry("minimumSize", m_currentSize);
    if (formFactor() == Plasma::Vertical) {
        m_currentSize.expandedTo(QSize(0, DEFAULT_PANEL_SIZE));
    } else {
        m_currentSize.expandedTo(QSize(DEFAULT_PANEL_SIZE, 0));
    }

    
    m_currentSize.expandedTo(QSize(screenRect.width(),DEFAULT_PANEL_SIZE));
    m_layout->setMaximumSize(QSize(screenRect.width(),DEFAULT_PANEL_SIZE));
    m_layout->setMinimumSize(QSize(screenRect.width(),DEFAULT_PANEL_SIZE));

    setMinimumSize(cg.readEntry("minimumSize", m_currentSize));
    setMaximumSize(cg.readEntry("maximumSize", m_currentSize));

    //connect(m_configureAction, SIGNAL(triggered()), this, SLOT(configurationInterface()));
    //END
    
}

void Panel::addPRosa(const char *path)
{
    Plasma::Applet *rosa_applet;
    rosa_applet = Applet::loadPlasmoid(path);
    if(rosa_applet){
      //kDebug() << "prog ROSA init";
      //printf("[*] Rosa panel Layout applet \n");
	rosa_applet->setParent(this); 
	//rosa_applet->setParentItem(this);
	layoutApplet(rosa_applet,QPointF(-1,-1));
	//m_layout->addItem(rosa_applet);
	rosa_applet->init();
	rosa_applet->updateConstraints(Plasma::StartupCompletedConstraint); 
	rosa_applet->flushPendingConstraintsEvents();
	rosa_applet->updateConstraints(Plasma::AllConstraints);
	rosa_applet->flushPendingConstraintsEvents();
	//layoutApplet(rosa_applet,QPointF(-1,-1));
    }    
    connect(rosa_applet, SIGNAL(sizeHintChanged(Qt::SizeHint)), this, SLOT(delayedUpdateSize()), Qt::UniqueConnection);
}

void Panel::addARosa(const char *name)
{
    Plasma::Applet *rosa_applet;
    rosa_applet = Applet::load(name);//Plasmoid("/usr/share/apps/rosapanel/brouser-applet");//("trash");
    if(rosa_applet){
      //kDebug() << "prog ROSA init";
      //printf("[*] Rosa panel Layout applet \n");
	rosa_applet->setParent(this); 
	//rosa_applet->setParentItem(this);
	layoutApplet(rosa_applet,QPointF(-1,-1));
	//m_layout->addItem(rosa_applet);
	rosa_applet->init();
	rosa_applet->updateConstraints(Plasma::StartupCompletedConstraint); 
	rosa_applet->flushPendingConstraintsEvents();
	rosa_applet->updateConstraints(Plasma::AllConstraints);
	rosa_applet->flushPendingConstraintsEvents();
	rosa_applet->setAspectRatioMode(Plasma::Square);
	//rosa_applet->containment()->enableAction("remove", false);
	//layoutApplet(rosa_applet,QPointF(-1,-1));
    }    
    connect(rosa_applet, SIGNAL(sizeHintChanged(Qt::SizeHint)), this, SLOT(delayedUpdateSize()), Qt::UniqueConnection);
}

void Panel::initRosa()
{
    printf("[*]Rosa initRosa!\n");

    KConfigGroup cg = config("Configuration");
    QString isInit("");
    isInit = cg.readEntry("RosaIsInit");    

    if(isInit == "7"){
	return;
    }

    cg.writeEntry("RosaIsInit","7");

    QRect screenRect = QDesktopWidget().screenGeometry();
    
    setLocation(Plasma::BottomEdge);
    resize(screenRect.width(),DEFAULT_PANEL_SIZE);
    
    setMaximumSize(QSize(screenRect.width(),DEFAULT_PANEL_SIZE));
    setMinimumSize(QSize(screenRect.width(),DEFAULT_PANEL_SIZE));
    
    m_currentSize.expandedTo(QSize(screenRect.width(),DEFAULT_PANEL_SIZE));
    m_layout->setMaximumSize(QSize(screenRect.width(),DEFAULT_PANEL_SIZE));
    m_layout->setMinimumSize(QSize(screenRect.width(),DEFAULT_PANEL_SIZE));

    updateSize();
    
    if(Applet::load("rosastarter")){
    	addApplet("rosastarter");
    }else{
	addApplet("launcher");
    }
    
    QVariantList args;
    
    args.append( KUrl("file:///usr/share/applications/kde4/dolphin.desktop").url() );
    addApplet("rosaicon",args);
    args.removeAt(0);
    args.append( KUrl("file:///usr/share/applications/firefox.desktop").url() );
    addApplet("rosaicon",args);
    args.removeAt(0);
    args.append( KUrl("file:///usr/share/applications/kde4/kopete.desktop").url() );
    addApplet("rosaicon",args);
    args.removeAt(0);
    args.append( KUrl("file:///usr/share/applications/mandriva-mozilla-thunderbird.desktop").url() );
    addApplet("rosaicon",args);
    args.removeAt(0);
    args.append( KUrl("file:///usr/share/applications/clementine.desktop").url() );
    addApplet("rosaicon",args);
    args.removeAt(0);
    args.append( KUrl("file:///usr/share/applications/kde4/systemsettings.desktop").url() );
    addApplet("rosaicon",args);
    args.removeAt(0);

    if(Plasma::Theme::defaultTheme()->themeName().toLower()=="rosa")
	    addApplet("rosaseparator");

    addARosa("tasks-applet");

    if(Plasma::Theme::defaultTheme()->themeName().toLower()=="rosa")
	    addApplet("rosaseparator");

    args.append( QDir::homePath() + "/" + i18n("Documents") );
    addApplet("stackfolder",args);
    args.removeAt(0);

    args.append(QDir::homePath() + "/" + i18n("Downloads"));
    addApplet("stackfolder",args);
    args.removeAt(0);
    
    addARosa("systemtray");   
    addARosa("digital-clock");
    addARosa("rosatrashcan");
    
    //enableAction("remove", false);
    
    //QStringList lsst = containmentActionsTriggers();
    //foreach (const QString &str, lsst){
    //	    printf("kkk %s\n",str.toAscii().constData());
    //}
    
    updateSize();
    
    saveContents(cg);

    printf("[*]Listing\n");

    for (int order = 0; order < m_layout->count(); ++order) {
        const Applet *applet = dynamic_cast<Applet *>(m_layout->itemAt(order));
        if (applet) {
            printf("[*] Applet %i name %s \n", order, QString::number(applet->id()).toAscii().constData());
        }
    }

    
}


QList<QAction*> Panel::contextualActions()
{
/* Analog of plasma configure panel. Realization in configpanel.h
    if (!m_configureAction) {
        m_configureAction = new QAction(i18n("ROSA Panel Settings"), this);
        m_configureAction->setIcon(KIcon("configure"));
        connect(m_configureAction, SIGNAL(triggered()), this, SLOT(configurationInterface()));

        constraintsEvent(Plasma::ImmutableConstraint);
    }

    QList<QAction*> actions;

    actions.append(m_configureAction);

    return actions;
*/
    QList<QAction*> actions;
    return actions;
}

void Panel::createConfigurationInterface(KConfigDialog *parent)
{

/*
    QWidget *mainConfiguration = new QWidget();
    mainConfiguration->setFixedSize(500,500);
    mainConfiguration->show();
*/
/*Now will fix bug on middle zone here*/


}

void Panel::configurationInterface()
{

    QRect screenRect = QDesktopWidget().screenGeometry();
    
    if(!m_mainConfiguration){
        ConfigPanel *mainConfiguration = new ConfigPanel();//ControllerWindow(0);
        mainConfiguration->setPanel(this);
	/*m_mainConfiguration->setFixedSize(screenRect.width(),50);
        m_mainConfiguration->move(0,screenRect.height() - 50);
	m_mainConfiguration->setWindowFlags(Qt::FramelessWindowHint | Qt::WindowStaysOnTopHint);
        m_mainConfiguration->setAcceptDrops(true);
    
	QBoxLayout *m_clayout = new QBoxLayout(QBoxLayout::LeftToRight);
	m_clayout->setContentsMargins(0, 0, 0, 0); 
    
    
	m_sizeTool = new QToolButton(m_mainConfiguration);
	m_sizeTool->setText("Height 0");
	//m_sizeTool->etIcon(icon);
	//m_sizeTool->setCheckable(false);
	//m_sizeTool->setAutoExclusive(false);
	//m_sizeTool->installEventFilter(this);
	m_sizeTool->setCursor(Qt::SizeVerCursor);
	m_clayout->addWidget(m_sizeTool);
	m_clayout->addStretch();
    
	m_mainConfiguration->setLayout(m_clayout);
	*/
	mainConfiguration->show();
	mainConfiguration->move(0, screenRect.height()-size().height()-50);

    }
}


void Panel::backgroundChanged()
{
    constraintsEvent(Plasma::LocationConstraint);
    update();
}

void Panel::adjustLastSpace()
{
    if (!m_layout) {
        return;
    }

    bool useSpacer = true;

    if (formFactor() == Plasma::Vertical) {
        foreach (Applet *applet, applets()) {
            if (applet->sizePolicy().verticalPolicy() & QSizePolicy::ExpandFlag) {
                useSpacer = false;
                break;
            }
        }
    } else {
        foreach (Applet *applet, applets()) {
            if (applet->sizePolicy().horizontalPolicy() & QSizePolicy::ExpandFlag) {
                useSpacer = false;
                break;
            }
        }
    }
/*
    if (useSpacer) {
        if (!m_lastSpace) {
            m_lastSpace = new Spacer(this);
            m_lastSpace->panel = this;
            m_lastSpace->m_visible = false;
            m_lastSpace->setPreferredSize(0,0);
            m_lastSpace->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
            m_layout->addItem(m_lastSpace);
        }
    } else {
        m_layout->removeItem(m_lastSpace);
        delete m_lastSpace;
        m_lastSpace = 0;
    }
*/    
}

void Panel::enableUpdateSize()
{
    m_canResize = true;
    if (m_resizedApplets > 0) {
        updateSize();
    }
}

void Panel::layoutApplet(Plasma::Applet* applet, const QPointF &pos)
{
    // this gets called whenever an applet is added, and we add it to our layout
    if (!m_layout) {
        return;
    }

    Plasma::FormFactor f = formFactor();
    int insertIndex = -1;

    //Enlarge the panel if possible and needed
    QSizeF appletHint = applet->preferredSize();
    QSizeF panelHint = m_layout->preferredSize();
    if (f == Plasma::Horizontal) {
        if (panelHint.width() + appletHint.width() > size().width()) {
            resize(panelHint.width() + appletHint.width(), size().height());
        }
    } else {
        if (panelHint.height() + appletHint.height() > size().height()) {
            resize(size().width(), panelHint.height() + appletHint.height());
        }
    }

    m_layout->setMinimumSize(size());
    m_layout->setMaximumSize(size());

    //if pos is (-1,-1) insert at the end of the panel
    if (pos != QPoint(-1, -1)) {
        for (int i = 0; i < m_layout->count(); ++i) {
            QRectF siblingGeometry = m_layout->itemAt(i)->geometry();
            if (f == Plasma::Horizontal) {
                qreal middle = (siblingGeometry.left() + siblingGeometry.right()) / 2.0;
                if (QApplication::layoutDirection() == Qt::RightToLeft) {
                    if (pos.x() > middle) {
                        insertIndex = i;
                    } else if (pos.x() >= siblingGeometry.left()) {
                        insertIndex = i + 1;
                    }
                } else if (pos.x() < middle) {
                    insertIndex = i;
                    break;
                } else if (pos.x() <= siblingGeometry.right()) {
                    insertIndex = i + 1;
                    break;
                }
            } else { // Plasma::Vertical
                qreal middle = (siblingGeometry.top() + siblingGeometry.bottom()) / 2.0;
                if (pos.y() < middle) {
                    insertIndex = i;
                    break;
                } else if (pos.y() <= siblingGeometry.bottom()) {
                    insertIndex = i + 1;
                    break;
                }
            }
        }
    }

/*
    const Applet *appletTest = dynamic_cast<Applet *>(m_layout->itemAt(insertIndex));
    if (appletTest) {
        if( appletTest->pluginName() == "tasks-applet" || (dynamic_cast<Applet *>(m_layout->itemAt(insertIndex-1)))->pluginName() == "tasks-applet" || (dynamic_cast<Applet *>(m_layout->itemAt(insertIndex+1)))->pluginName() == "tasks-applet" ){

	    //foreach (Applet *applet, applets()) {
    		//if (applet->pluginName() == "icon") {
    	    	applet->setMaximumWidth(0);
		applet->setMaximumHeight(0);
		applet->hide();
    		//}
	    //}
	    return;
    	}
    }
*/

    if(applet->pluginName() == "icon"){
	applet->setMaximumWidth(0);
	applet->setMaximumHeight(0);
	KConfigGroup acg;
	acg = applet->config();
	QString url = acg.readEntry("Url");
	QVariantList args;
	args.append( KUrl(url).url() );
	//applet->load("rosaicon", applet->id(),args);
	QSizeF p = applet->size();
	applet->close();
	applet->hide();
	corona()->removeItem(applet);
	applet->destroy();
	addApplet("rosaicon", args, QRectF(pos,p));
	KConfigGroup cg = config("Configuration");
	saveContents(cg);
	return;
    }
/*
    foreach (Applet *app, applets()) {

        if (app) {
            if(app->pluginName() == "icon"){
            	    //corona()->removeItem(applet);
		    //applet->destroy();
		    app->setMaximumWidth(0);
		    app->setMaximumHeight(0);
		    app->hide();
	    }
        }

    }
*/
    m_layout->removeItem(m_lastSpace);

    if (insertIndex == -1 || insertIndex >= m_layout->count()) {
        m_layout->addItem(applet);
    } else {
        m_layout->insertItem(insertIndex, applet);
    }

    if (m_lastSpace) {
        m_layout->addItem(m_lastSpace);
    }

    //FIXME: there must be some beter way to do this rather than this rather error prone arbitrary wait
    m_lastSpaceTimer->start(2000);

    connect(applet, SIGNAL(sizeHintChanged(Qt::SizeHint)), this, SLOT(delayedUpdateSize()), Qt::UniqueConnection);
}

void Panel::delayedUpdateSize()
{
    ++m_resizedApplets;

    if (!m_updateSizeTimer->isActive()) {
        m_updateSizeTimer->start();
    }
}

void Panel::appletWasRemoved(Plasma::Applet* applet)
{

/*
    if(applet->name() == "tasks-applet"){
	return;
    }
*/
    disconnect(applet, SIGNAL(sizeHintChanged(Qt::SizeHint)), this, SLOT(delayedUpdateSize()));

    if (!m_layout) {
        return;
    }

    m_layout->removeItem(applet);

    //shrink the panel if possible
    if (formFactor() == Plasma::Horizontal) {
        resize(size().width() - applet->size().width(), size().height());
    } else {
        resize(size().width(), size().height() - applet->size().height());
    }

    m_layout->setMaximumSize(size());
    m_lastSpaceTimer->start(200);
}

void Panel::updateSize()
{
  //kDebug() << "Update size called: " << m_currentSize;

  //kDebug() << "Can resize: " << m_canResize;
  //kDebug() << "Need resize: " << m_resizedApplets;

  
  //setMinimumHeight(24);
  //setMinimumWidth(24);
  
  //if(formFactor() == Plasma::Vertical)
  //{
      //kDebug() << "We are Vertical!";
      //kDebug() << "Layout size: " << m_layout->size();
  //}
  
    if (!m_canResize || m_resizedApplets < 1) {
        m_resizedApplets = 0;
        return;
    }

    foreach (Applet *applet, applets()) {
      //applet->flushPendingConstraintsEvents();
      //applet->updateConstraints(Plasma::AllConstraints);
      //applet->flushPendingConstraintsEvents();
      //QRectF origGeometry = applet->geometry();
      //applet->setGeometry(QRectF(0,0,0,0));
      //applet->setGeometry(origGeometry);
      applet->update();
    };
    
    m_resizedApplets = 0;
    m_canResize = false;

    const bool horizontal = formFactor() != Plasma::Vertical;
    int delta = horizontal ? size().width() : size().height();
    //kDebug() << "Iterating applets: ";
    foreach (Applet *applet, applets()) {
      //kDebug() << "Applet size: " << applet->size();
        if (horizontal) {
            delta -= applet->preferredSize().width();
        } else {
            delta -= applet->preferredSize().height();
        }
    }
    delta *= -1;

    //setting the preferred width when delta = 0 and preferredWidth() < minimumWidth()
    // leads to the same thing as setPreferredWidth(minimumWidth())

    if (delta != 0) {
        // amazing but true: preferedSize doesn't take into consideration margins.
        qreal l, t, r, b;
        m_layout->getContentsMargins(&l, &t, &r, &b);

        if (horizontal) {
            QRect screenRect = QDesktopWidget().screenGeometry();
	    setPreferredWidth(screenRect.width());
            //setPreferredWidth(preferredWidth() + delta + l + r);
        } else {
            setPreferredHeight(preferredHeight() + delta + t + b);
        }
    }


    //kDebug() << "resize to" << preferredSize() << delta << ", was" << size();
    resize(preferredSize());

    //for a while we won't execute updateSize() again
    m_enableUpdateResizeTimer->start();
    
    /*QRect screenRect = QDesktopWidget().screenGeometry();
    
    m_currentSize.expandedTo(QSize(screenRect.width(),DEFAULT_PANEL_SIZE));
    m_layout->setMaximumSize(QSize(screenRect.width(),DEFAULT_PANEL_SIZE));
    m_layout->setMinimumSize(QSize(screenRect.width(),DEFAULT_PANEL_SIZE));
    */
    
}

void Panel::updateBorders(const QRect &geom, bool inPaintEvent)
{
    Plasma::Location loc = location();
    FrameSvg::EnabledBorders enabledBorders = FrameSvg::AllBorders;

    int s = screen();
    //kDebug() << loc << s << formFactor() << geometry();

    qreal topHeight = m_background->marginSize(Plasma::TopMargin);
    qreal bottomHeight = m_background->marginSize(Plasma::BottomMargin);
    qreal leftWidth = m_background->marginSize(Plasma::LeftMargin);
    qreal rightWidth = m_background->marginSize(Plasma::RightMargin);

    //remove unwanted borders
    if (s < 0) {
        // do nothing in this case, we want all the borders
    } else if (loc == BottomEdge || loc == TopEdge) {
        QRect r = QDesktopWidget().screenGeometry();//Kephal::ScreenUtils::screenGeometry(s);

        if (loc == BottomEdge) {
            enabledBorders ^= FrameSvg::BottomBorder;
            bottomHeight = 0;
        } else {
            enabledBorders ^= FrameSvg::TopBorder;
            topHeight = 0;
        }

        if (geom.x() <= r.x()) {
            enabledBorders ^= FrameSvg::LeftBorder;
            leftWidth = 0;
        }
        if (geom.right() >= r.right()) {
            enabledBorders ^= FrameSvg::RightBorder;
            rightWidth = 0;
        }

        //kDebug() << "top/bottom: Width:" << width << ", height:" << height;
    } else if (loc == LeftEdge || loc == RightEdge) {
        QRect r = QDesktopWidget().screenGeometry();//Kephal::ScreenUtils::screenGeometry(s);

        if (loc == RightEdge) {
            enabledBorders ^= FrameSvg::RightBorder;
            rightWidth = 0;
        } else {
            enabledBorders ^= FrameSvg::LeftBorder;
            leftWidth = 0;
        }

        if (geom.y() <= r.y()) {
            enabledBorders ^= FrameSvg::TopBorder;
            topHeight = 0;
        }

        if (geom.bottom() >= r.bottom()) {
            enabledBorders ^= FrameSvg::BottomBorder;
            bottomHeight = 0;
        }

        //kDebug() << "left/right: Width:" << width << ", height:" << height;
    } else {
      //kDebug() << "no location!?";
    }

    //activate borders and fetch sizes again
    m_background->setEnabledBorders(enabledBorders);
    m_background->getMargins(leftWidth, topHeight, rightWidth, bottomHeight);

    //calculation of extra margins has to be done after getMargins
    const QGraphicsItem *box = toolBox();
    if (box && immutability() == Mutable) {
        QSizeF s = box->boundingRect().size();
        if (formFactor() == Vertical) {
            //hardcoded extra margin for the toolbox right now
            bottomHeight += s.height() + 2;
            //Default to horizontal for now
        } else {
            rightWidth += s.width() + 2;
        }
    }

    //invalidate the layout and set again
    if (m_layout) {
        switch (location()) {
        case LeftEdge:
            rightWidth = qMin(rightWidth, qMax(qreal(2), size().width() - KIconLoader::SizeMedium));
            break;
        case RightEdge:
            leftWidth = qMin(leftWidth, qMax(qreal(2), size().width() - KIconLoader::SizeMedium));
            break;
        case TopEdge:
            bottomHeight = qMin(bottomHeight, qMax(qreal(2), size().height() - KIconLoader::SizeMedium));
            break;
        case BottomEdge:
            topHeight = qMin(topHeight, qMax(qreal(2), size().height() - KIconLoader::SizeMedium));
            break;
        default:
            break;
        }

        m_layout->setContentsMargins(leftWidth, topHeight, rightWidth, bottomHeight);

        if (!inPaintEvent) {
            resize(preferredSize());
        }
    }
    
    /*QRect screenRect = QDesktopWidget().screenGeometry();
    
    m_currentSize.expandedTo(QSize(screenRect.width(),DEFAULT_PANEL_SIZE));
    m_layout->setMaximumSize(QSize(screenRect.width(),DEFAULT_PANEL_SIZE));
    m_layout->setMinimumSize(QSize(screenRect.width(),DEFAULT_PANEL_SIZE));
    */
}

void Panel::constraintsEvent(Plasma::Constraints constraints)
{
    if (constraints & Plasma::FormFactorConstraint) {
        m_maskDirty = true;

        Plasma::FormFactor form = formFactor();
        Qt::Orientation layoutDirection = form == Plasma::Vertical ? Qt::Vertical : Qt::Horizontal;
        // create or set up our layout!
        if (m_layout) {
            m_layout->setMaximumSize(size());
            m_layout->setOrientation(layoutDirection);
        }
    }
/*
    foreach (Applet *app, applets()) {

        if (app) {
            if(app->pluginName() == "icon"){
            	    //corona()->removeItem(applet);
		    //applet->destroy();
		    app->setMaximumWidth(0);
		    app->setMaximumHeight(0);
		    app->hide();
	    }
        }

    }
*/
    //we need to know if the width or height is 100%
    if (constraints & Plasma::LocationConstraint || constraints & Plasma::SizeConstraint) {
        m_maskDirty = true;
        m_currentSize = geometry().size().toSize();
        QRectF screenRect = screen() >= 0 ? QDesktopWidget().screenGeometry()/*Kephal::ScreenUtils::screenGeometry(screen())*/ :
                                            geometry();

        if ((formFactor() == Horizontal && m_currentSize.width() >= screenRect.width()) ||
            (formFactor() == Vertical && m_currentSize.height() >= screenRect.height())) {
            m_background->setElementPrefix(location());
        } else {
            switch (location()) {
            case LeftEdge:
                //this call will automatically fallback at no prefix if the element isn't available
                m_background->setElementPrefix("west-mini");
                break;
            case RightEdge:
                m_background->setElementPrefix("east-mini");
                break;
            case TopEdge:
                m_background->setElementPrefix("north-mini");
                break;
            case BottomEdge:
            default:
                m_background->setElementPrefix("south-mini");
                break;
            }
        }

        m_background->resizeFrame(m_currentSize);

        //FIXME: this seems the only way to correctly resize the layout the first time when the
        // saved panel size is less than the default is to setting a maximum size.
        // this shouldn't happen. maybe even a qgraphicslayout bug?
        if (m_layout && (constraints & Plasma::SizeConstraint)) {
            m_layout->setMaximumSize(size());
        }

        if (constraints & Plasma::LocationConstraint) {
            setFormFactorFromLocation(location());
        }
    }

    if (constraints & Plasma::StartupCompletedConstraint) {
        connect(this, SIGNAL(appletAdded(Plasma::Applet*,QPointF)),
                this, SLOT(layoutApplet(Plasma::Applet*,QPointF)));
    }

    if (constraints & Plasma::ImmutableConstraint) {
        bool unlocked = immutability() == Plasma::Mutable;

        if (m_configureAction) {
            m_configureAction->setEnabled(unlocked);
            m_configureAction->setVisible(unlocked);
        }

        m_maskDirty = true;
        updateBorders(geometry().toRect());
    }
    
    /*if(location()==Plasma::TopEdge){
	setLocation(Plasma::BottomEdge);
    }*/
}

void Panel::saveState(KConfigGroup &config) const
{
    config.writeEntry("minimumSize", minimumSize());
    config.writeEntry("maximumSize", maximumSize());
}

void Panel::paintInterface(QPainter *painter,
                           const QStyleOptionGraphicsItem *option,
                           const QRect& contentsRect)
{
    Q_UNUSED(contentsRect)
    //FIXME: this background drawing is bad and ugly =)
    // draw the background untransformed (saves lots of per-pixel-math)
    painter->resetTransform();

    const Containment::StyleOption *containmentOpt = qstyleoption_cast<const Containment::StyleOption *>(option);

    QRect viewGeom;
    if (containmentOpt && containmentOpt->view) {
        viewGeom = containmentOpt->view->geometry();
    } else {
        viewGeom = m_lastViewGeom;
    }

    if (m_maskDirty || m_lastViewGeom != viewGeom) {
        m_maskDirty = false;
        m_lastViewGeom = viewGeom;

        updateBorders(viewGeom, true);
        if (containmentOpt && containmentOpt->view && !m_background->mask().isEmpty()) {
            const QRegion mask = m_background->mask();
            containmentOpt->view->setMask(mask);
            //Plasma::WindowEffects::enableBlurBehind(containmentOpt->view->winId(), true, mask);
        }
    }

    // blit the background (saves all the per-pixel-products that blending does)
    painter->setCompositionMode(QPainter::CompositionMode_Source);
    painter->setRenderHint(QPainter::Antialiasing);

    m_background->paintFrame(painter, option->exposedRect);
}

void Panel::setFormFactorFromLocation(Plasma::Location loc) {
    printf("[*] Set FF\n");
    switch (loc) {
        case BottomEdge:
        case TopEdge:
            //kDebug() << "setting horizontal form factor";
            setFormFactor(Plasma::Horizontal);
            printf("[*] Horizont\n");
            break;
        case RightEdge:
        case LeftEdge:
            //kDebug() << "setting vertical form factor";
            setFormFactor(Plasma::Vertical);
            break;
        case Floating:
            //TODO: implement a form factor for floating panels
            kDebug() << "Floating is unimplemented.";
            break;
        default:
            kDebug() << "invalid location!!";
    }
}

void Panel::showDropZone(const QPoint pos)
{
    if (!scene() || !m_layout) {
        return;
    }

    if (pos == QPoint()) {
        if (m_spacer) {
            m_layout->removeItem(m_spacer);
            m_spacer->hide();
        }
        return;
    }

    //lucky case: the spacer is already in the right position
    if (m_spacer && m_spacer->geometry().contains(pos)) {
        return;
    }

    Plasma::FormFactor f = formFactor();
    int insertIndex = m_layout->count();

    //FIXME: needed in two places, make it a function?
    for (int i = 0; i < m_layout->count(); ++i) {
        QRectF siblingGeometry = m_layout->itemAt(i)->geometry();

        if (f == Plasma::Horizontal) {
            qreal middle = siblingGeometry.left() + (siblingGeometry.width() / 2.0);
            if (pos.x() < middle) {
                insertIndex = i;
                break;
            } else if (pos.x() <= siblingGeometry.right()) {
                insertIndex = i + 1;
                break;
            }
        } else { // Plasma::Vertical
            qreal middle = siblingGeometry.top() + (siblingGeometry.height() / 2.0);
            if (pos.y() < middle) {
                insertIndex = i;
                break;
            } else if (pos.y() <= siblingGeometry.bottom()) {
                insertIndex = i + 1;
                break;
            }
        }
    }
/*
    const Applet *applet = dynamic_cast<Applet *>(m_layout->itemAt(insertIndex));

    if (applet) {
        if(applet->pluginName() == "tasks-applet" || (dynamic_cast<Applet *>(m_layout->itemAt(insertIndex-1)))->pluginName() == "tasks-applet" || (dynamic_cast<Applet *>(m_layout->itemAt(insertIndex+1)))->pluginName() == "tasks-applet" ){
	    return;
    	}
    }
*/

    m_spacerIndex = insertIndex;
    if (insertIndex != -1) {
        if (!m_spacer) {
            m_spacer = new Spacer(this);
            m_spacer->panel = this;
        } else {
            m_layout->removeItem(m_spacer);
        }
    	m_spacer->show();
    	m_layout->insertItem(insertIndex, m_spacer);
    }
}

void Panel::restore(KConfigGroup &group)
{
    Containment::restore(group);

    //kDebug() << "[*] ROSA start restore\n";

    KConfigGroup appletsConfig(&group, "Applets");

    QMap<int, Applet *> oderedApplets;
    QList<Applet *> unoderedApplets;
    QList<Applet *> undeltedApplets;
    bool mistake = false;

    foreach (Applet *applet, applets()) {
        KConfigGroup appletConfig(&appletsConfig, QString::number(applet->id()));
        KConfigGroup layoutConfig(&appletConfig, "LayoutInformation");

        int order = layoutConfig.readEntry("Order", -1);
	
	
	if( mistake == false ){
    	    if (order > -1 ) {
    		if(applet->name() == "rosaicon"){
		    KUrl url (appletConfig.readEntry("Url", ""));
    		    if (!url.protocol().isEmpty()) {
            		    url = KIO::NetAccess::mostLocalUrl(url, 0);
    		    }
		    if(KDesktopFile::isDesktopFile(url.toLocalFile())){
        		oderedApplets[order] = applet;
        	    }else{
        		unoderedApplets.append(applet);
        	    }
        	}else if(applet->name() != "rosaicon"){
        	    oderedApplets[order] = applet;
        	}else{
		    applet->setMaximumWidth(0);
		    applet->setMaximumHeight(0);
        	    layoutConfig.writeEntry("Order", "-1");
        	    layoutConfig.writeEntry("BadApplet", "1");
        	    layoutConfig.writeEntry("geometry", "0,0,0,0");

        	    unoderedApplets.append(applet);
        	}
    	    //if LayoutInformation is not available use the usual way, as a bonus makes it retrocompatible with oler configs
    	    } else {
        	unoderedApplets.append(applet);
    	    }
    	}else{
    	    undeltedApplets.append(applet);
    	}

        connect(applet, SIGNAL(sizeHintChanged(Qt::SizeHint)), this, SLOT(delayedUpdateSize()), Qt::UniqueConnection);
        if(applet->pluginName() == "trash"){
		mistake = true;
	}
    }

    foreach (Applet *applet, oderedApplets) {
	if(applet->name() == "launcher"){
	    if(Applet::load("rosastarter")){
		addARosa("rosastarter");
		continue;
	    }
	}
        if (m_lastSpace) {
            m_layout->insertItem(m_layout->count()-1, applet);
        } else {
            m_layout->addItem(applet);
        }
        if(applet->pluginName() == "trash") {
		break;
	}
    }

    foreach (Applet *applet, unoderedApplets) {
	if(applet->name() == "launcher"){
	    if(Applet::load("rosastarter")){
		addARosa("rosastarter");
		continue;
	    }
	}
        //layoutApplet(applet, applet->pos());
        if(applet->pluginName() == "trash"){
		break;
	}
    }
    
    //foreach (Applet *applet, undeltedApplets) {
    //}
    foreach (Applet *applet, unoderedApplets) {    
	    applet->close();
	    applet->hide();
	    applet->destroy();
	    corona()->removeItem(applet);
    }
    
    mistake = false;
    Plasma::Applet::List pleftApplets = containment()->applets();
    foreach(Plasma::Applet* applet, pleftApplets){
	if(mistake){
	    applet->close();
	    applet->hide();
	    applet->destroy();
	    corona()->removeItem(applet);
	}
	if(applet->pluginName() == "trash"){
		mistake = true;
	}
    }

    updateSize();
}

void Panel::saveContents(KConfigGroup &group) const
{
    Containment::saveContents(group);

    KConfigGroup appletsConfig(&group, "Applets");
    for (int order = 0; order < m_layout->count(); ++order) {
        const Applet *applet = dynamic_cast<Applet *>(m_layout->itemAt(order));
        if (applet) {
            if(applet->pluginName() == "icon"){
            	    //applet->close();
		    //applet->hide();
		    //corona()->removeItem(applet);
		    //applet->destroy();
		    //applet->setMaximumWidth(0);
		    //applet->setMaximumHeight(0);
    		    KConfigGroup appletConfig(&appletsConfig, QString::number(applet->id()));
        	    KConfigGroup layoutConfig(&appletConfig, "LayoutInformation");

        	    layoutConfig.writeEntry("Order", "-1");
        	    layoutConfig.writeEntry("BadApplet", "1");
        	    layoutConfig.writeEntry("geometry", "0,0,0,0");
		    break;
	    }

            KConfigGroup appletConfig(&appletsConfig, QString::number(applet->id()));
            KConfigGroup layoutConfig(&appletConfig, "LayoutInformation");

            layoutConfig.writeEntry("Order", order);
            if(applet->pluginName() == "trash"){
		    break;
	    }
        }
    }
    //KConfigGroup cg = config("Configuration");
    //cg.writeEntry("RosaIsInit","7");
}

K_EXPORT_PLASMA_APPLET(panel, Panel)

#include "panel.moc"

